@@ -1,9 +1,11 @@
-import validators
+from datetime import datetime
 
+import validators
 from ._odata_filter import _compose_filter
 from ._str_match_algo import fuzzy_match
 from .guardpoint_dataclasses import Cardholder, SortAlgorithm
 from .guardpoint_error import GuardPointError
+from .guardpoint_utils import GuardPointResponse
 
 
 class CardholdersAPI:
@@ -16,6 +18,9 @@ class CardholdersAPI:
         url_query_params = "(" + cardholder.uid + ")"
 
         code, json_body = self.gp_json_query("DELETE", url=(url + url_query_params))
+        # Check response body is formatted correctly
+        if json_body:
+            GuardPointResponse.check_odata_body_structure(json_body)
 
         if code != 204:  # HTTP NO_CONTENT
             try:
@@ -40,9 +45,19 @@ class CardholdersAPI:
             if len(cardholder.cardholderPersonalDetail.changed_attributes) > 0:
                 self.update_personal_details(cardholder.uid, cardholder.cardholderPersonalDetail)
 
+        if cardholder.cards:
+            if isinstance(cardholder.cards, list):
+                for card in cardholder.cards:
+                    if len(card.changed_attributes) > 0:
+                        if validators.uuid(card.uid):
+                            self.update_card(card)
+                        else:
+                            card.cardholderUID = cardholder.uid
+                            self.new_card(card)
+
         ch = cardholder.dict(editable_only=True, changed_only=True)
 
-        if len(ch) < 1: # Nothing to update
+        if len(ch) < 1:  # Nothing to update
             return True
 
         url = "/odata/API_Cardholders"
@@ -51,12 +66,13 @@ class CardholdersAPI:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            #'IgnoreNonEditable': ''
+            # 'IgnoreNonEditable': ''
         }
 
-
-
         code, json_body = self.gp_json_query("PATCH", headers=headers, url=(url + url_query_params), json_body=ch)
+        # Check response body is formatted correctly
+        if json_body:
+            GuardPointResponse.check_odata_body_structure(json_body)
 
         if code != 204:  # HTTP NO_CONTENT
             if 'error' in json_body:
@@ -70,7 +86,7 @@ class CardholdersAPI:
 
     def new_card_holder(self, cardholder: Cardholder, overwrite_cardholder=False):
 
-        #url = "/odata/API_Cardholders/CreateFullCardholder"
+        # url = "/odata/API_Cardholders/CreateFullCardholder"
         url = "/odata/API_Cardholders"
 
         headers = {
@@ -79,18 +95,31 @@ class CardholdersAPI:
             'IgnoreNonEditable': ''
         }
 
-        ch = cardholder.dict(editable_only=True)
+        ch = cardholder.dict(editable_only=True, changed_only=True)
 
         # When using CreateFullCardholder
-        #body = {'cardholder': ch}
+        # body = {'cardholder': ch}
         # print(json.dumps(body))
         code, json_body = self.gp_json_query("POST", headers=headers, url=url, json_body=ch)
+        # Check response body is formatted correctly
+        if json_body:
+            GuardPointResponse.check_odata_body_structure(json_body)
 
         if code == 201:  # HTTP CREATED
-            return Cardholder(json_body)
+            new_cardholder = Cardholder(json_body)
+            if cardholder.cardholderPersonalDetail:
+                self.update_personal_details(cardholder_uid=new_cardholder.uid,
+                                             personal_details=cardholder.cardholderPersonalDetail)
+            if cardholder.cardholderCustomizedField:
+                self.update_custom_fields(cardholder_uid=new_cardholder.uid,
+                                          customFields=cardholder.cardholderCustomizedField)
+
+            return self._get_card_holder(new_cardholder.uid)
+
         elif code == 422:  # unprocessable Entity
             if "errorMessages" in json_body:
-                raise GuardPointError(f'{json_body["errorMessages"][0]["message"]}-{json_body["errorMessages"][0]["other"]}')
+                raise GuardPointError(
+                    f'{json_body["errorMessages"][0]["message"]}-{json_body["errorMessages"][0]["other"]}')
         else:
             if "errorMessages" in json_body:
                 raise GuardPointError(json_body["errorMessages"][0]["other"])
@@ -123,6 +152,9 @@ class CardholdersAPI:
                                        "insideArea"
 
         code, json_body = self.gp_json_query("GET", url=(url + url_query_params))
+        # Check response body is formatted correctly
+        if json_body:
+            GuardPointResponse.check_odata_body_structure(json_body)
 
         if code != 200:
             if isinstance(json_body, dict):
@@ -131,23 +163,20 @@ class CardholdersAPI:
             else:
                 raise GuardPointError(str(code))
 
-        # Check response body is formatted as expected
-        if not isinstance(json_body, dict):
-            raise GuardPointError("Badly formatted response.")
-        if 'value' not in json_body:
-            raise GuardPointError("Badly formatted response.")
-
         return Cardholder(json_body['value'][0])
 
     def get_card_holders(self, offset: int = 0, limit: int = 10, search_terms: str = None, areas: list = None,
                          filter_expired: bool = False, cardholder_type_name: str = None,
                          sort_algorithm: SortAlgorithm = SortAlgorithm.SERVER_DEFAULT, threshold: int = 75,
-                         count = False):
+                         count: bool = False, earliest_last_pass: datetime = None):
 
         url = "/odata/API_Cardholders"
 
-        filter_str = _compose_filter(search_words=search_terms, areas=areas, filter_expired=filter_expired,
-                                     cardholder_type_name=cardholder_type_name)
+        filter_str = _compose_filter(search_words=search_terms,
+                                     areas=areas,
+                                     filter_expired=filter_expired,
+                                     cardholder_type_name=cardholder_type_name,
+                                     earliest_last_pass=earliest_last_pass)
 
         url_query_params = ("?" + filter_str)
 
@@ -155,28 +184,25 @@ class CardholdersAPI:
             url_query_params += "$count=true&$top=0"
         else:
             url_query_params += "$expand=" \
-                                    "cardholderType($select=typeName)," \
-                                    "cards," \
-                                    "cardholderPersonalDetail," \
-                                    "cardholderCustomizedField," \
-                                    "insideArea," \
-                                    "securityGroup&" \
-                                    "$orderby=fromDateValid%20desc&"
+                                "cardholderType($select=typeName)," \
+                                "cards," \
+                                "cardholderPersonalDetail," \
+                                "cardholderCustomizedField," \
+                                "insideArea," \
+                                "securityGroup&" \
+                                "$orderby=fromDateValid%20desc&"
             url_query_params += "$top=" + str(limit) + "&$skip=" + str(offset)
 
         code, json_body = self.gp_json_query("GET", url=(url + url_query_params))
+        # Check response body is formatted correctly
+        if json_body:
+            GuardPointResponse.check_odata_body_structure(json_body)
 
         if code != 200:
-            if isinstance(json_body, dict):
-                if 'error' in json_body:
-                    raise GuardPointError(json_body['error'])
-
-        if not isinstance(json_body, dict):
-            raise GuardPointError("Badly formatted response.")
-        if 'value' not in json_body:
-            raise GuardPointError("Badly formatted response.")
-        if not isinstance(json_body['value'], list):
-            raise GuardPointError("Badly formatted response.")
+            if 'error' in json_body:
+                raise GuardPointError(f"{json_body['error']} - ({code})")
+            else:
+                raise GuardPointError(f"No error message - ({code})")
 
         if count:
             return json_body['@odata.count']
