@@ -6,16 +6,12 @@ from enum import Enum
 log = logging.getLogger(__name__)
 
 
-class SortAlgorithm(Enum):
-    SERVER_DEFAULT = 0,
-    FUZZY_MATCH = 1
-
-
 class Observable:
+    # A set of all attributes which get changed
+    changed_attributes = set()
+
     def __init__(self):
         self.observed = defaultdict(list)
-        # A set of all attributes which get changed
-        self.changed_attributes = set()
 
     def __setattr__(self, name, value):
         super().__setattr__(name, value)
@@ -27,39 +23,59 @@ class Observable:
         self.observed[name].append(lambda name: self.changed_attributes.add(name))
 
 
+def sanitise_args(obj: Observable, args, kwargs):
+    kwarg_dict = dict()
+
+    for arg in args:
+        if isinstance(arg, dict):
+            kwarg_dict.update(arg)
+
+    for k, v in kwargs.items():
+        if hasattr(type(obj), k):
+            kwarg_dict[k] = v
+            obj.changed_attributes.add(k)
+        else:
+            log.debug(f"{obj.__class__.__name__}.{k} - attribute ignored.")
+            # raise ValueError(f"No such attribute: {k}")
+
+    return kwarg_dict
+
+
+class SortAlgorithm(Enum):
+    SERVER_DEFAULT = 0,
+    FUZZY_MATCH = 1
+
 
 @dataclass
-class Card:
+class Card(Observable):
     technologyType: int = 0
     description: str = ""
     cardCode: str = ""
     status: str = "Free"
     cardholderUID: any = None
     cardType: str = "Magnetic"
-    readerFunctionUID: str = ""
+    readerFunctionUID: any = None
     uid: str = ""
 
-    def __init__(self, card_dict: dict):
+    def __init__(self, *args, **kwargs):
+        super().__init__()
+        card_dict = sanitise_args(self, args, kwargs)
+
         for property_name in card_dict:
-            if isinstance(card_dict[property_name], str):
+            if isinstance(card_dict[property_name], (str, type(None), bool, int)):
                 setattr(self, property_name, card_dict[property_name])
+                self.add_observer(property_name)
 
-            if isinstance(card_dict[property_name], type(None)):
-                setattr(self, property_name, None)
+    def _remove_non_changed(self, ch: dict):
+        for key, value in list(ch.items()):
+            if key not in self.changed_attributes:
+                ch.pop(key)
+        return ch
 
-            if isinstance(card_dict[property_name], bool):
-                setattr(self, property_name, bool(card_dict[property_name]))
-
-    def dict(self, editable_only=False):
+    def dict(self, editable_only=False, changed_only=False):
         c = {}
         for k, v in asdict(self).items():
-            if isinstance(v, list):
-                c[k] = v
-            elif isinstance(v, dict):
-                c[k] = v
-            elif isinstance(v, bool):
-                c[k] = v
-            elif isinstance(v, int):
+            if isinstance(v, (list, dict, bool, int)):
                 c[k] = v
             elif isinstance(v, type(None)):
                 c[k] = None
@@ -69,6 +85,11 @@ class Card:
         if editable_only:
             if 'uid' in c:
                 c.pop('uid')
+            if 'readerFunctionUID' in c:
+                c.pop('readerFunctionUID')
+
+        if changed_only:
+            c = self._remove_non_changed(c)
 
         return c
 
@@ -80,15 +101,9 @@ class Area:
 
     def __init__(self, area_dict: dict):
         for property_name in area_dict:
-            if isinstance(area_dict[property_name], str):
+            if isinstance(area_dict[property_name], (str, type(None), bool, int)):
                 setattr(self, property_name, area_dict[property_name])
 
-            if isinstance(area_dict[property_name], type(None)):
-                setattr(self, property_name, None)
-
-            if isinstance(area_dict[property_name], bool):
-                setattr(self, property_name, bool(area_dict[property_name]))
-
     def dict(self):
         return {k: str(v) for k, v in asdict(self).items()}
 
@@ -104,60 +119,37 @@ class SecurityGroup:
 
     def __init__(self, security_group_dict: dict):
         for property_name in security_group_dict:
-            if isinstance(security_group_dict[property_name], str):
+            if isinstance(security_group_dict[property_name], (str, type(None), bool, int)):
                 setattr(self, property_name, security_group_dict[property_name])
 
-            if isinstance(security_group_dict[property_name], type(None)):
-                setattr(self, property_name, None)
-
-            if isinstance(security_group_dict[property_name], bool):
-                setattr(self, property_name, bool(security_group_dict[property_name]))
-
     def dict(self):
         return {k: str(v) for k, v in asdict(self).items()}
 
+
 @dataclass
-class ScheduledMag:
+class ScheduledMag(Observable):
     uid: str = ""
     securityGroupAPIKey: str = ""
     scheduledSecurityGroupUID: str = ""
     cardholderUID: str = ""
-    toDateValid:str = ""
-    fromDateValid:str = ""
+    toDateValid: str = ""
+    fromDateValid: str = ""
     status: str = ""
 
     def __init__(self, *args, **kwargs):
-        scheduled_mags_dict = dict()
-        for arg in args:
-            if isinstance(arg, dict):
-                scheduled_mags_dict = arg
-
-        for k,v in kwargs.items():
-            scheduled_mags_dict[k] = v
-
         super().__init__()
+        scheduled_mags_dict = sanitise_args(self, args, kwargs)
 
-        #Initialise clss attributes from dictionary
+        # Initialise clss attributes from dictionary
         for property_name in scheduled_mags_dict:
-            if isinstance(scheduled_mags_dict[property_name], str):
+            if isinstance(scheduled_mags_dict[property_name], (str, type(None), bool, int)):
                 setattr(self, property_name, scheduled_mags_dict[property_name])
-
-            if isinstance(scheduled_mags_dict[property_name], type(None)):
-                setattr(self, property_name, None)
-
-            if isinstance(scheduled_mags_dict[property_name], bool):
-                setattr(self, property_name, bool(scheduled_mags_dict[property_name]))
+                self.add_observer(property_name)
 
     def dict(self, editable_only=False):
         c = {}
         for k, v in asdict(self).items():
-            if isinstance(v, list):
-                c[k] = v
-            elif isinstance(v, dict):
-                c[k] = v
-            elif isinstance(v, bool):
-                c[k] = v
-            elif isinstance(v, int):
+            if isinstance(v, (list, dict, bool, int)):
                 c[k] = v
             elif isinstance(v, type(None)):
                 c[k] = None
@@ -171,6 +163,8 @@ class ScheduledMag:
                 c.pop('status')
 
         return c
+
+
 @dataclass
 class CardholderCustomizedField(Observable):
     uid: str = ""
@@ -210,29 +204,19 @@ class CardholderCustomizedField(Observable):
     cF_StringField_19: str = ""
     cF_StringField_20: str = ""
 
-    def __init__(self, custom_fields_dict: dict):
+    def __init__(self, *args, **kwargs):
         super().__init__()
+        custom_fields_dict = sanitise_args(self, args, kwargs)
+
         for property_name in custom_fields_dict:
-            if isinstance(custom_fields_dict[property_name], str):
+            if isinstance(custom_fields_dict[property_name], (str, type(None), bool, int)):
                 setattr(self, property_name, custom_fields_dict[property_name])
                 self.add_observer(property_name)
 
-            if isinstance(custom_fields_dict[property_name], type(None)):
-                setattr(self, property_name, None)
-                self.add_observer(property_name)
-
-            if isinstance(custom_fields_dict[property_name], bool):
-                setattr(self, property_name, bool(custom_fields_dict[property_name]))
-                self.add_observer(property_name)
-
     def dict(self, changed_only=False):
         c = dict()
         for k, v in asdict(self).items():
-            if isinstance(v, list):
-                c[k] = v
-            elif isinstance(v, dict):
-                c[k] = v
-            elif isinstance(v, bool):
+            if isinstance(v, (bool, int)):
                 c[k] = v
             elif isinstance(v, int):
                 c[k] = v
@@ -255,41 +239,32 @@ class CardholderCustomizedField(Observable):
 
 @dataclass
 class CardholderPersonalDetail(Observable):
-    officePhone: any
-    cityOrDistrict: str
-    streetOrApartment: str
-    postCode: str
-    privatePhoneOrFax: str
-    mobile: str
-    email: str
-    carRegistrationNum: str
-    company: str
-    idFreeText: str
+    uid: str = ""
+    officePhone: str = ""
+    cityOrDistrict: str = ""
+    streetOrApartment: str = ""
+    postCode: str = ""
+    privatePhoneOrFax: str = ""
+    mobile: str = ""
+    email: str = ""
+    carRegistrationNum: str = ""
+    company: str = ""
+    idFreeText: str = ""
     idType: str = "IdentityCard"
 
-    def __init__(self, person_details_dict: dict):
+    def __init__(self, *args, **kwargs):
         super().__init__()
-        for property_name in person_details_dict:
-            if isinstance(person_details_dict[property_name], str):
-                setattr(self, property_name, person_details_dict[property_name])
-                self.add_observer(property_name)
+        person_details_dict = sanitise_args(self, args, kwargs)
 
-            if isinstance(person_details_dict[property_name], type(None)):
-                setattr(self, property_name, None)
+        for property_name in person_details_dict:
+            if isinstance(person_details_dict[property_name], (str, type(None), bool, int)):
+                setattr(self, property_name, person_details_dict[property_name])
                 self.add_observer(property_name)
 
     def dict(self, changed_only=False):
         ch_pd = dict()
         for k, v in asdict(self).items():
-            if isinstance(v, list):
-                ch_pd[k] = v
-            elif isinstance(v, dict):
-                ch_pd[k] = v
-            elif isinstance(v, bool):
-                ch_pd[k] = v
-            elif isinstance(v, int):
-                ch_pd[k] = v
-            elif isinstance(v, type(None)):
+            if isinstance(v, type(None)):
                 ch_pd[k] = None
             else:
                 ch_pd[k] = str(v)
@@ -331,12 +306,12 @@ class Cardholder(Observable):
     cardholderPersonalDetail: CardholderPersonalDetail = None
     cardholderCustomizedField: CardholderCustomizedField = None
     insideArea: Area = None
-    ownerSiteUID: any = None
+    ownerSiteUID: str = ""
     securityGroupApiKey: any = None
     ownerSiteApiKey: any = None
     accessGroupApiKeys: any = None
     liftAccessGroupApiKeys: any = None
-    cardholderTypeUID: str = "11111111-1111-1111-1111-111111111111"
+    cardholderTypeUID: str = ""
     departmentUID: any = None
     description: str = ""
     grantAccessForSupervisor: bool = False
@@ -344,33 +319,26 @@ class Cardholder(Observable):
     needEscort: bool = False
     personalWeeklyProgramUID: any = None
     pinCode: str = ""
-    sharedStatus: any = None
-    securityGroupUID: any = None
+    sharedStatus: str = ""
+    securityGroupUID: str = ""
     accessGroupUIDs: any = None
     liftAccessGroupUIDs: any = None
     lastDownloadTime: any = None
-    lastInOutArea: any = None
+    lastInOutArea: str = ""
     lastInOutReaderUID: any = None
     lastInOutDate: any = None
     lastAreaReaderDate: any = None
     lastAreaReaderUID: any = None
     lastPassDate: any = None
     lastReaderPassUID: any = None
-    insideAreaUID: any = None
+    insideAreaUID: str = ""
     cards: list = None
 
     def __init__(self, *args, **kwargs):
-        cardholder_dict = dict()
-        for arg in args:
-            if isinstance(arg, dict):
-                cardholder_dict = arg
-
-        for k, v in kwargs.items():
-            cardholder_dict[k] = v
         super().__init__()
+        cardholder_dict = sanitise_args(self, args, kwargs)
 
         for property_name in cardholder_dict:
-            # If we have a list - For example, a cardholder has many cards - we only take the first entry
             if isinstance(cardholder_dict[property_name], list):
                 if property_name == "cards":
                     setattr(self, property_name, [])
@@ -379,6 +347,14 @@ class Cardholder(Observable):
                 else:
                     setattr(self, property_name, cardholder_dict[property_name])
 
+            if property_name == "cardholderPersonalDetail":
+                if isinstance(cardholder_dict[property_name], CardholderPersonalDetail):
+                    self.cardholderPersonalDetail = cardholder_dict[property_name]
+
+            if property_name == "cardholderCustomizedField":
+                if isinstance(cardholder_dict[property_name], CardholderCustomizedField):
+                    self.cardholderCustomizedField = cardholder_dict[property_name]
+
             if isinstance(cardholder_dict[property_name], dict):
                 if property_name == "insideArea":
                     self.insideArea = Area(cardholder_dict[property_name])
@@ -391,18 +367,10 @@ class Cardholder(Observable):
                 if property_name == "cardholderCustomizedField":
                     self.cardholderCustomizedField = CardholderCustomizedField(cardholder_dict[property_name])
 
-            if isinstance(cardholder_dict[property_name], str):
+            if isinstance(cardholder_dict[property_name], (str, type(None), bool, int)):
                 setattr(self, property_name, cardholder_dict[property_name])
                 self.add_observer(property_name)
 
-            if isinstance(cardholder_dict[property_name], type(None)):
-                setattr(self, property_name, None)
-                self.add_observer(property_name)
-
-            if isinstance(cardholder_dict[property_name], bool):
-                setattr(self, property_name, bool(cardholder_dict[property_name]))
-                self.add_observer(property_name)
-
     def to_search_pattern(self):
         pattern = ""
         if self.firstName:
@@ -429,11 +397,7 @@ class Cardholder(Observable):
     def dict(self, editable_only=False, changed_only=False):
         ch = dict()
         for k, v in asdict(self).items():
-            if isinstance(v, list):
-                ch[k] = v
-            elif isinstance(v, dict):
-                ch[k] = v
-            elif isinstance(v, bool):
+            if isinstance(v, (list, dict, bool, int)):
                 ch[k] = v
             elif isinstance(v, type(None)):
                 ch[k] = None
